"""Examples to the HierMat package

In this module are example scripts to demonstrate the usage of the package

model_1d module
------------------

.. automodule:: HierMat.examples.model_1d
    :members:

"""